#include <stdio.h>
#include <stdlib.h>

#include "vcl.h"
#include "vrt.h"

#include "vcc_if.h"
#include "parse_encoding.h"
#include "wb.h"

/*
 * SAFE LENGTH ESTIMATES
 */

#define base64_decode_l(l)	((l * 3 / 4) + 1)
#define base64_encode_l(l)	(((((l * 4 / 3) + 3) / 4) * 4) + 1)
#define hex_decode_l(l)		((l * 2) + 1)
#define hex_encode_l(l)		((l / 2) + 1)

static const struct vmod_priv const null_blob[1] =
{
	{
		.priv = "",
		.len = 0,
		.free = NULL
	}
};

struct b64_state_s {
	uint32_t	magic;
#define B64_STATE_MAGIC	0xc0b64b64
	unsigned char	bytes[2];
	ssize_t	l;
};

static struct b64_alphabet {
	char *b64;
	char i64[256];
	int padding;
} b64_alphabet[__MAX_ENCODING];

static void
alpha_init(struct b64_alphabet *alpha)
{
	int i;
	const char *p;

	for (i = 0; i < 256; i++)
		alpha->i64[i] = -1;
	for (p = alpha->b64, i = 0; *p; p++, i++)
		alpha->i64[(int)*p] = (char)i;
	if (alpha->padding)
		alpha->i64[alpha->padding] = 0;
}

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e event)
{
	(void) priv;

	if (event != VCL_EVENT_LOAD)
		return 0;

	b64_alphabet[BASE64].b64 =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		"ghijklmnopqrstuvwxyz0123456789+/";
	b64_alphabet[BASE64].padding = '=';
	b64_alphabet[BASE64URL].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URL].padding = '=';
	b64_alphabet[BASE64URLNOPAD].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URLNOPAD].padding = 0;
	alpha_init(&b64_alphabet[BASE64]);
	alpha_init(&b64_alphabet[BASE64URL]);
	alpha_init(&b64_alphabet[BASE64URLNOPAD]);
	return (0);
}

/*
 * Base64-encode *in (size: inlen) into *out, max outlen bytes. If there is
 * insufficient space, it will bail out and return -1. Otherwise, it will
 * null-terminate and return the used space.
 * The alphabet `a` defines... the alphabet. Padding is optional.
 * Inspired heavily by gnulib/Simon Josefsson (as referenced in RFC4648)
 *
 * XXX: tmp[] and idx are used to ensure the reader (and author) retains
 * XXX: a limited amount of sanity. They are strictly speaking not
 * XXX: necessary, if you don't mind going crazy.
 *
 * FIXME: outlenorig is silly. Flip the logic.
 */
static VCL_BLOB
base64_encode (struct ws *ws, const enum encoding enc, struct vmod_priv *rblob,
    const char *in, size_t inlen)
{
	struct b64_alphabet *alpha = &b64_alphabet[enc];
	struct wb_s wb;
	char *p;
	size_t outlen;
	unsigned char tmp[3], idx;

	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
	AN(rblob);

	if (in == NULL || inlen == 0)
		return null_blob;

	p = wb_create(ws, &wb);
	outlen = wb_space(&wb);
	if (outlen < base64_encode_l(inlen))
		return NULL;

	while (1) {
		assert(inlen);
		assert(outlen > 3);

		tmp[0] = (unsigned char) in[0];
		tmp[1] = (unsigned char) in[1];
		tmp[2] = (unsigned char) in[2];

		*p++ = alpha->b64[(tmp[0] >> 2) & 0x3f];

		idx = (tmp[0] << 4);
		if (inlen > 1)
			idx += (tmp[1] >> 4);
		idx &= 0x3f;
		*p++ = alpha->b64[idx];

		if (inlen > 1) {
			idx = (tmp[1] << 2);
			if (inlen > 2)
				idx += tmp[2] >> 6;
			idx &= 0x3f;

			*p++ = alpha->b64[idx];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		if (inlen > 2) {
			*p++ = alpha->b64[tmp[2] & 0x3f];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		/*
		 * XXX: Only consume 4 bytes, but since we need a fifth for
		 * XXX: NULL later on, we might as well test here.
		 */
		assert(outlen >= 5);

		outlen -= 4;

		if (inlen < 4)
			break;

		inlen -= 3;
		in += 3;
	}

	assert(outlen > 0);
	wb_advanceP(&wb, p);
	(void) wb_finish_blob(&wb, rblob);
	if (rblob->len == 0)
		return null_blob;
	return rblob;
}



static inline char
hexnibblelc(char c)
{
	return ((c > 9) ? ('a' + (c - 0xa)) : ('0' + c));
}


/*
 * en/decoders need to work for the string_list vargs case (implicit length) as
 * well as for the blob case (explicit length). Fortunately, the blob case is
 * not varadic, so we can assume just one length argument _or_ varags
 *
 * rblob: alloc additional whitespace so any result can also be (safely)
 *	  used as a string.
 */

static VCL_BLOB
hex_encode_va(struct ws *ws, struct vmod_priv *rblob, const ssize_t len,
    const char *in, va_list va)
{
	struct wb_s wb[1];
	char *p;
	ssize_t insz;
	int i;

	AN(rblob);

	if (in == vrt_magic_string_end)
		return null_blob;

	p = wb_create(ws, wb);
	if (p == NULL)
		return NULL;
	const char *e = wb_end(wb);

	if (len != -1)
		insz = len;
	else if (in == NULL)
		insz = 0;
	else
		insz = strlen(in);

	do {
		if (insz == 0)
			goto next;
		for (i = 0; i < insz && p + 1 < e; i++) {
			*p++ = hexnibblelc((in[i] & 0xf0) >> 4);
			*p++ = hexnibblelc( in[i] & 0x0f);
		}
		if (i != insz)
			goto err;

	  next:
		in = va_arg(va, const char *);
		if (in == vrt_magic_string_end)
			break;
		insz = (in != NULL) ? strlen(in) : 0;
	} while(1);

	wb_advanceP(wb, p);

	(void) wb_finish_blob(wb, rblob);

	if (rblob->len == 0)
		return null_blob;

	return rblob;

  err:
	wb_reset(wb);
	return NULL;
}

static VCL_BLOB
hex_encode(struct ws *ws, struct vmod_priv *rblob, const ssize_t len,
    const char *in, ...)
{
	va_list ap;
	VCL_BLOB r;
	va_start(ap, in);
	r = hex_encode_va(ws, rblob, len, in, ap);
	va_end(ap);
	return r;
}


/* Functions */

VCL_BLOB
vmod_decode(VRT_CTX, VCL_ENUM decs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);

	return null_blob;
}

VCL_STRING
vmod_encode(VRT_CTX, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding enc = parse_encoding(encs);
	struct vmod_priv tmp[1];
	const struct vmod_priv *r = NULL;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(b);

	switch(enc) {
	case IDENTITY:
		// XXX
		break;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		r = base64_encode(ctx->ws, enc, tmp, b->priv, b->len);
		break;
	case HEX:
	case HEXLC: // XXX: lower-case?
		r = hex_encode(ctx->ws, tmp, b->len, b->priv,
		    vrt_magic_string_end);
		break;
	default:
		WRONG("Illegal encoding");
	}

	if (r == NULL)
		return NULL;
	return r->priv;
}

VCL_STRING __match_proto__(td_convert_transcode)
vmod_transcode(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	struct vmod_priv tmp[1];

	va_list ap;
	VCL_BLOB r;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);
	assert(enc != _INVALID);

	// XXX DO THE DECODE BIT

	assert(enc == HEXLC);	// XXX
	va_start(ap, p);
	r = hex_encode_va(ctx->ws, tmp, -1, p, ap);
	va_end(ap);
	if (r == NULL)
		return NULL;
	return r->priv;
}
VCL_BLOB
vmod_transcode_blob(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);
	assert(enc != _INVALID);
	return b;
}
